/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/afs/protocol/querymanager.c
 * @brief forwarding of queries
 * @author Christian Grothoff
 *
 * The query manager is responsible for queueing queries.  Queued
 * queries are used to fill buffers (instead of using noise). The QM
 * is also responsible for selecting the initial set of nodes that
 * will receive the query. For a good choice, it keeps track of which
 * nodes were recently hot in answering queries. Some randomness is
 * preserved to ensure that we potentially find a better path.<p>
 *
 * Routing is an incredibly hard problem, so please consider
 * consulting with other gnunet-developers before making any
 * significant changes here, even if you have CVS write access.
 **/

#include "config.h"

#include <string.h>

#include "util/checksum.h"
#include "util/logging.h"
#include "util/returnvalues.h"
#include "util/xmalloc.h"
#include "applications/afs/protocol/afs.h"
#include "applications/afs/protocol/querymanager.h"

/* default size of the bitmap: 16 byte = 128 bit */
#define BITMAP_SIZE 16

/* of how many outbound queries do we simultaneously
   keep track? */
#define QUERY_RECORD_COUNT 512

#define DEBUG_QUERYMANAGER NO 

/**
 * In this struct, we store information about a 
 * query that is being send from the local node to
 * optimize the sending strategy. 
 **/
typedef struct {

  /**
   * How often did we send this query so far?
   **/
  unsigned int sendCount;  
  
  /**
   * The message that we are sending.
   **/
  AFS_p2p_QUERY * msg;

  /**
   * Bit-map marking the hostIndices (computeIndex)
   * of nodes that have received this query already.
   * Note that the bit-map has a maximum size, if
   * the index is out-of-bounds, it is hashed into
   * the smaller size of the bitmap. There may thus be
   * nodes with identical indices, in that case, only one of
   * the nodes will receive the query.
   **/
  unsigned char bitmap[BITMAP_SIZE];

  /**
   * When do we stop forwarding (!) this query?
   **/
  cron_t expires;

  /**
   * How many nodes were connected when we
   * initated sending this query?
   **/
  unsigned int activeConnections;

  /**
   * What is the total distance of the query to
   * the connected nodes?
   **/
  unsigned long long totalDistance;

} QueryRecord;

/**
 * Which are the most productive nodes that we are
 * currently connected to? In this array, we count
 * the number of replies we have recently received
 * from every node. The values are devided by 2 on
 * a regular basis to reflect changes over time.
 **/
static unsigned int recordNodes[BITMAP_SIZE*8];

/**
 * Sum of all entries in recordNodes.
 **/
static unsigned int totalRecordCount;

/**
 * Array of the queries we are currently sending out.
 **/
static QueryRecord queries[QUERY_RECORD_COUNT];

/**
 * Mutex for all query manager structures.
 **/
static Mutex * queryManagerLock;

/**
 * Map the id to an index into the recordNodes or bitmap array.
 **/
static int getIndex(const HostIdentity * id) {
  unsigned int index;

  index = coreAPI->computeIndex(id);
  if (index > 8*BITMAP_SIZE)
    index = index & (8*BITMAP_SIZE-1);
  return index;
}

static void setBit(QueryRecord * qr,
		   int bit) {
  unsigned char theBit = (1 << (bit & 7));
  qr->bitmap[bit>>3] |= theBit;
}

static int getBit(QueryRecord * qr,
		  int bit) {
  unsigned char theBit = (1 << (bit & 7));
  return (qr->bitmap[bit>>3] & theBit) > 0;
}

/**
 * Callback method for filling buffers. This method
 * is invoked by the core if a message is about to
 * be send and there is space left for a 3QUERY.
 * We then search the pending queries and fill one
 * (or more) in if possible.
 * @param receiver the receiver of the message
 * @param position is the reference to the
 *        first unused position in the buffer where GNUnet is building
 *        the message
 * @param padding is the number of bytes left in that buffer.
 * @return the number of bytes written to
 *   that buffer (must be a positive number).
 **/
static int fillInQuery(HostIdentity * receiver,
		       void * position,
		       int padding) {
  static unsigned int pos = 0;
  unsigned int start;
  unsigned int delta;
  cron_t now;

  cronTime(&now);
  MUTEX_LOCK(queryManagerLock);
  start = pos;
  delta = 0;
  while (padding - delta > sizeof(AFS_p2p_QUERY)+sizeof(HashCode160)) {
    if ( (queries[pos].expires > now) &&
	 (getBit(&queries[pos],
		 getIndex(receiver)) == 0) &&
	 (padding - delta >=
	  ntohs(queries[pos].msg->header.size) ) ) {
      HexName qhex;
      HexName hhex;

      IFLOG(LOG_DEBUG,
	    hash2hex(&receiver->hashPubKey,
		     &hhex);
	    hash2hex(&queries[pos].msg->queries[0],
		     &qhex));
#if DEBUG_QUERYMANAGER
      LOG(LOG_DEBUG,
	  "DEBUG: adding %d queries (%s) to outbound buffer of %s\n",
	  (ntohs(queries[pos].msg->header.size)-sizeof(AFS_p2p_QUERY))/sizeof(HashCode160),
	  &qhex,
	  &hhex);
#endif
      setBit(&queries[pos],
	     getIndex(receiver));
      memcpy(&((char*)position)[delta],
	     queries[pos].msg,
	     ntohs(queries[pos].msg->header.size));
      queries[pos].sendCount++;
      delta += ntohs(queries[pos].msg->header.size);
    }
    pos++;
    if (pos >= QUERY_RECORD_COUNT)
      pos = 0;
    if (pos == start)
      break;
  }
  MUTEX_UNLOCK(queryManagerLock);
  return delta;
}

/**
 * Decrease the record counters by half.
 **/
static void decreaseRecordCounters(void * unused) {
  int i;

  if (totalRecordCount < BITMAP_SIZE)
    return; /* insignificantly small numbers, do not reduce! */
  MUTEX_LOCK(queryManagerLock);
  totalRecordCount = 0;
  for (i=0;i<BITMAP_SIZE*8;i++)
    totalRecordCount += (recordNodes[i] /= 2);
  MUTEX_UNLOCK(queryManagerLock);
}

/**
 * We got content back from a node, remember that!
 **/
void noticeContentReceived(const HostIdentity * sender) {
  if (sender == NULL)
    return;
  MUTEX_LOCK(queryManagerLock);  
  recordNodes[getIndex(sender)]++;
  MUTEX_UNLOCK(queryManagerLock);
}

/**
 * Initialize the query management.
 **/
int initQueryManager() {
  int i;

  for (i=0;i<QUERY_RECORD_COUNT;i++) {
    queries[i].expires = 0; /* all expired */
    queries[i].msg = NULL;
  }
  for (i=0;i<BITMAP_SIZE*8;i++)
    recordNodes[i] = 0; /* no replies received so far */
  totalRecordCount = 0;
  queryManagerLock = coreAPI->getConnectionModuleLock();
  addCronJob(&decreaseRecordCounters,
	     5 * cronMINUTES,
	     5 * cronMINUTES,
	     NULL); 
  coreAPI->registerSendCallback(sizeof(AFS_p2p_QUERY)+sizeof(HashCode160),
				&fillInQuery);
  return OK;
}

void doneQueryManager() {
  int i;

  for (i=0;i<QUERY_RECORD_COUNT;i++) 
    FREENONNULL(queries[i].msg);
  delCronJob(&decreaseRecordCounters,
	     5 * cronMINUTES,
	     NULL);
  coreAPI->unregisterSendCallback(sizeof(AFS_p2p_QUERY)+sizeof(HashCode160),
				  &fillInQuery);
}

/**
 * A "PerNodeCallback" method that selects the most
 * active nodes for forwarding (with some randomness).
 *
 * Also computes the sum of the distances of the
 * other node IDs as a side-effect.
 **/
static void selectActiveNodes(HostIdentity * id,
			      QueryRecord * qr) {
  HexName hex;

  IFLOG(LOG_EVERYTHING,
	hash2hex(&id->hashPubKey,
		 &hex));
  LOG(LOG_EVERYTHING,
      "EVERYTHING: selecting from active nodes %s as rand(%d)*3 > rand(%d)\n",
      &hex, 
      recordNodes[getIndex(id)]+1,
      totalRecordCount+1);
  /* Forward the query to on average 3 nodes.
     Prefer recently active nodes (proportionally). */
  if (randomi(recordNodes[getIndex(id)]+1) >
      randomi(totalRecordCount/3+1)) {
    setBit(qr, getIndex(id));
    LOG(LOG_EVERYTHING,
	"EVERYTHING: node selected for forwarding due to activity: %s\n",
	&hex);
  } else
    qr->totalDistance += distanceHashCode160(&qr->msg->queries[0],
					     &id->hashPubKey);
}

/**
 * A "PerNodeCallback" method that selects some
 * random nodes (biased according to proximity).
 **/
static void selectRandomNodes(HostIdentity * id,
			      QueryRecord * qr) {
  unsigned int avDist;
  HexName hex;

  if ( (qr->totalDistance == 0) ||
       (qr->activeConnections == 0) ) {
    return; /* activeConnections should never be 0,
	       if totalDistance is 0, this is caused
	       by us selectAcitveNodes selecting
	       all nodes already */
  }
  avDist = qr->totalDistance / qr->activeConnections;

  IFLOG(LOG_EVERYTHING,
	hash2hex(&id->hashPubKey,
		 &hex));
  LOG(LOG_EVERYTHING,
      "EVERYTHING: selecting at random from active nodes %s using rand(%d)*3 > rand(%d)\n",
      &hex,
      distanceHashCode160(&qr->msg->queries[0],
			  &id->hashPubKey) / qr->activeConnections,
      avDist);

  /* Select 3 random nodes for forwarding. Give preference  
     to nodes that are close.  Division by 4 to ensure
     that we are in the range of a signed int. */
  if (randomi(1+distanceHashCode160(&qr->msg->queries[0],
				    &id->hashPubKey) / qr->activeConnections / 4)
      > randomi(1+avDist/12)) {
    LOG(LOG_EVERYTHING,
	"EVERYTHING: node selected for forwarding from random set: %s\n",
	&hex);
    setBit(qr, getIndex(id));
  }
}

/**
 * A "PerNodeCallback" method that forwards
 * the query to the selected nodes.
 **/
static void sendToSelected(HostIdentity * id,
			   QueryRecord * qr) {
  if (getBit(qr, getIndex(id)) == 1) {
    HexName hex;

    IFLOG(LOG_EVERYTHING,
	  hash2hex(&id->hashPubKey,
		   &hex));
    LOG(LOG_EVERYTHING,
	"EVERYTHING: queueing query in buffer of selected node %s\n",
	&hex);
    coreAPI->sendToNode(id,
			&qr->msg->header,
			ntohl(qr->msg->priority),
			TTL_DECREMENT);
  }
}

/**
 * Take a query and forward it to the appropriate
 * number of nodes (depending on load, queue, etc).
 **/
void forwardQuery(AFS_p2p_QUERY * msg) {
  cron_t now;
  QueryRecord * qr;
  QueryRecord dummy;
  HexName hex;
  cron_t oldestTime;
  cron_t expirationTime;
  int oldestIndex;
  int i;
  
  IFLOG(LOG_DEBUG,
	hash2hex(&msg->queries[0],
		 &hex));
#if DEBUG_QUERYMANAGER
  LOG(LOG_DEBUG,
      "DEBUG: forwarding query for %s with ttl %d\n",
      &hex,
      ntohl(msg->ttl));
#endif
  cronTime(&now);
  MUTEX_LOCK(queryManagerLock);
  
  oldestIndex = -1;
  expirationTime = now + ntohl(msg->ttl) / 4;
  oldestTime = expirationTime;
  for (i=0;i<QUERY_RECORD_COUNT;i++)
    if (queries[i].expires < oldestTime) {
      oldestTime = queries[i].expires;
      oldestIndex = i;      
    }
  if (oldestIndex == -1) {				    
#if DEBUG_QUERYMANAGER
    LOG(LOG_DEBUG,
	"DEBUG: keeping track of %d queries already, will not manage this one\n",
	QUERY_RECORD_COUNT);    
#endif
    qr = &dummy;  
  } else {
    qr = &queries[oldestIndex];
    FREENONNULL(qr->msg);
    qr->msg = NULL;
  } 
  qr->expires = expirationTime;
  qr->msg = MALLOC(ntohs(msg->header.size));
  memcpy(qr->msg,
	 msg,
	 ntohs(msg->header.size));
  memset(&qr->bitmap[0],
	 0,
	 BITMAP_SIZE);
  qr->totalDistance = 0;
  qr->activeConnections
    = coreAPI->forAllConnectedNodes
    ((PerNodeCallback)&selectActiveNodes,
     qr);
  coreAPI->forAllConnectedNodes
    ((PerNodeCallback)&selectRandomNodes,
     qr);
  /* now forward to a couple of selected nodes */
  coreAPI->forAllConnectedNodes
    ((PerNodeCallback)&sendToSelected,
     qr);

  if (qr == &dummy)
    FREE(dummy.msg);
  MUTEX_UNLOCK(queryManagerLock);
}


/* end of querymanager.c */
